'use strict';

/*
TODO: Preview
TODO: Change the alert with something better
*/

(function() {
/**
 * The placeholder text to writer in the canvas to tell to put the file there
 */
const placeholderText = 'Drop the file'

/**
 * The PDF documents, populated once they are loaded
 */
var documents = {}

/**
 * The number of pages in documents
 */
var pageNumbers = {}

/**
 * The classes to use for li of the page list
 */
var listClasses = {
	'doc1': 'page-item badge badge-primary',
	'doc2': 'page-item badge badge-success'
}

/**
 * The references to original files that will be uploaded
 */
var originalFiles = {}

/**
 * Get the id of the document by traversing parents.
 *
 * In practice, go to the first ancestor (div, but not checked) that has an id.
 *
 * @param elem The element to traverse the parents of
 * @return The desired id, or null if it has not been found
 */
function getFirstId(elem) {
	var tmp = elem
	while (tmp.parentNode && !tmp.id) {
		tmp = tmp.parentNode
	}
	return tmp.id
}

/**
 * Initialize the canvas: set a text for the user and initialize the callbacks
 *
 * @param elem The element to initialize
 */
function initCanvas(elem) {
	canvasPlaceholder(elem)

	elem.addEventListener('drop', dropFile)
	elem.addEventListener('dragover', e => {
		e.preventDefault()
	})
	elem.addEventListener('click', canvasClick)
}

/**
 * Set the placeholder text on a canvas
 *
 * @param elem The canvas on which write the placeholder
 */
function canvasPlaceholder(elem) {
	// Reset responsive units and pass to pixels
	elem.className = 'default-canvas'
	var w = elem.clientWidth, h = elem.clientHeight
	elem.className = ''
	elem.width = w
	elem.height = h

	var ctx = elem.getContext('2d')
	ctx.font = '1.5rem sans-serif'
	ctx.textAlign = 'center'
	ctx.fillText(placeholderText, elem.clientWidth / 2, elem.clientHeight / 2)

	/* Since this function is called also in case of error, make sure that the
	navigation is hidden when this is visible */
	elem.parentNode.querySelector('.pdf-pages').style.display = 'none'
}

/**
 * Handle the drop of a file
 *
 * @param e The event
 */
function dropFile(e) {
	e.preventDefault()

	var file = null

	if (e.dataTransfer.items) {
		// We need only a file, but cycle, to reject non files
		for (var i = 0; i < e.dataTransfer.items.length; i++) {
			if (e.dataTransfer.items[i].kind === 'file') {
				file = e.dataTransfer.items[i].getAsFile()
				break
			}
		}
	} else if(e.dataTransfer.files.length > 0) {
		file = e.dataTransfer.files[0]
	}

	loadPdf(e.target, file)

	var id = getFirstId(e.target)
	var selected = document
			.querySelectorAll('.page-item[data-merge-id=' + id + ']')
	for (var i = 0; i < selected.length; i++) {
		var s = selected[i]
		s.parentNode.removeChild(s)
		checkPageList()
	}

	originalFiles[id] = file
}

/**
 * Load and start drawing the PDF on the canvas
 *
 * @param elem The element where draw the canvas into
 * @param file The file to draw on the canvas
 */
function loadPdf(elem, file) {
	var fr = new FileReader()

	/* Exploit the structure of the document: the parent of the canvas contain
	everything about that document, and so has the id doc1 or doc2 */
	var parent = elem.parentNode
	var id = parent.id
	documents[id] = null

	fr.onload = e => {
		pdfjsLib.getDocument(e.target.result)
		.then(pdf => {
			documents[id] = pdf
			pageNumbers[id] = 1

			pdf.getPage(1).then(page => renderPage(elem, page))

			parent.querySelector('.pdf-pages input').value = '1'
			parent.querySelector('.pdf-pages').style.display = 'flex'
		})
		.catch(e => {
			canvasPlaceholder(elem)
			alert(e.message)
		})
	}

	fr.readAsArrayBuffer(file)
}

function renderPage(canvas, page) {
	var context = canvas.getContext('2d')

	var viewport1 = page.getViewport(1)
	/* Fit to width by faking 100% initially.
	Notice that we have to use style because width does not support 100%. */
	canvas.style.width = '100%'
	var scaleW = canvas.clientWidth / viewport1.width
	canvas.style.width = ''
	var viewport = page.getViewport(scaleW)

	canvas.height = viewport.height
	context.canvas.height = viewport.height
	canvas.width = viewport.width
	context.canvas.width = viewport.width

	var renderContext = {
		canvasContext: context,
		viewport: viewport
	}

	page.render(renderContext)
}

/**
 * Handle the click event on a canvas
 *
 * @param e The click event
 */
function canvasClick(e) {
	var clickedId = getFirstId(e.target)

	if (documents[clickedId] == null) {
		return
	}

	listAddPage(clickedId, pageNumbers[clickedId])

	Object.keys(documents).forEach(doc => {
		changePage(doc, 1, true)
	})
}

/**
 * Change the page in a PDF
 *
 * @param doc The id of the document container in which the page should be
 * changed
 * @param page The number of page to set
 * @param relative Tell if the page is the number of page or a number of page to
 * forward/rewind
 * @return false if the page is not valid, true otherwise
 */
function changePage(doc, page, relative) {
	var container = document.getElementById(doc)
	if (container == null || documents[doc] == null) {
		return false
	}

	if (relative) {
		page += pageNumbers[doc]
	}

	if (page > documents[doc].pdfInfo.numPages || page < 1) {
		return false
	}

	pageNumbers[doc] = page
	container.querySelector('input').value = page

	documents[doc].getPage(page)
	.then(pdfPage => renderPage(container.querySelector('canvas'), pdfPage))

	return true
}

function buttonClicked(e) {
	var dir = 0,
		elem = e.target

	if (elem.className.indexOf('prev') != -1) {
		dir = -1
	} else if (elem.className.indexOf('next') != -1) {
		dir = 1
	} else {
		return
	}

	var id = getFirstId(elem)
	if (id != null) {
		changePage(id, dir, true)
	}
}

/**
 * Handle the change of text in the page input
 *
 * Notice that this event is caused only by user interaction, and not by
 * programmatic variations of the value attribute
 *
 * @param e The change event
 */
function inputChanged(e) {
	var id = getFirstId(e.target)
	if (id == null || documents[id] == null) {
		return
	}

	var newPage = parseInt(e.target.value)

	if (isNaN(newPage) || !changePage(id, newPage, false)) {
		e.target.value = pageNumbers[id]
	}
}

/**
 * Add a new page to the list
 *
 * @param doc The document to take the page from
 * @param page The page number
 */
function listAddPage(doc, page) {
	var li = document.createElement('li')
	li.className = listClasses[doc]
	li.title = 'Right click to remove, drag to move'
	li.setAttribute('data-merge-id', doc)
	li.setAttribute('data-merge-page', page)
	li.textContent = doc + '-' + page

	var parent = document.getElementById('page-list')
	parent.appendChild(li)

	sortable('#page-list', 'reload')

	li.addEventListener('click', e => {
		changePage(doc, page, false)
	})

	li.addEventListener('contextmenu', e => {
		e.preventDefault()
		parent.removeChild(li)
		checkPageList()

		sortable('#page-list', 'reload')
	})

	checkPageList()
}

/**
 * Check the list of pages, to decide wheter to show the no page message or the
 * merge button
 */
function checkPageList() {
	if (document.getElementById('page-list').childElementCount == 0) {
		document.getElementById('no-pages').style.display = 'block'
		document.getElementById('merge-container').style.display = 'none'
	} else {
		document.getElementById('no-pages').style.display = 'none'
		document.getElementById('merge-container').style.display = 'block'
	}
}

initCanvas(document.querySelector('#doc1 canvas'))
initCanvas(document.querySelector('#doc2 canvas'))

/* Button and inputs are only for PDF navigation, so we can go relaxed with
these selectors */
document.querySelectorAll('button').forEach(e =>
	e.addEventListener('click', buttonClicked)
)
document.querySelectorAll('input').forEach(e =>
	e.addEventListener('change', inputChanged)
)

document.getElementById('merge-button').addEventListener('click', e => {
	var formData = new FormData()
	var pages = ''

	Object.keys(originalFiles).forEach(id => {
		var file = originalFiles[id]
		formData.append(id, file, file.name)
	})

	document.querySelectorAll('#page-list li').forEach(li => {
		pages += li.getAttribute('data-merge-id') + '-' +
				li.getAttribute('data-merge-page') + ' '
	})
	formData.append('pages', pages)

	var xhr = new XMLHttpRequest()
	xhr.open('POST', 'upload', true)

	/* https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/Sending_and_Receiving_Binary_Data
	Creating the blob from the response did not work, where getting directly
	the blob from AJAX work as a charm. */
	xhr.responseType = 'blob'

	xhr.onload = function () {
		if (this.status == 200) {
			var link = document.createElement('a')
			link.href = window.URL.createObjectURL(xhr.response)
			link.download = 'merged.pdf'

			// The usual trick to download blobs
			document.body.appendChild(link)
			link.click()
			document.body.removeChild(link)
		} else {
			alert('Merge failed')
		}
	}

	xhr.send(formData)
})

sortable('#page-list', {
	placeholder: '<li class="badge badge-light placeholder">Move the page here</li>'
})

window.addEventListener('resize', e => {
	var canvases = document.getElementsByTagName('canvas')
	for (var i = 0; i < canvases.length; i++) {
		let id = getFirstId(canvases[i])
		if (documents[id] != null) {
			changePage(id, 0, true)
		} else {
			canvasPlaceholder(canvases[i])
		}
	}
})

})()
